import csv
import numpy as np
import scipy.optimize as so
import matplotlib.pyplot as plt
from mymodule_sup import *
import BlockAverage as ba

fin_name = input('Input data file(*.csv) = ')
fout_name = input('Output file name(*.txt) = ')
fout = open(fout_name, 'w')
fout.write('Input date file...{}\n'.format(fin_name))
r = 5           #  r-metric is used to estimate an approximate configuration

with open(fin_name, 'r') as fin:
    raw_data = [v for v in csv.reader(fin)]

obj_labels = raw_data[0][1:]
n_obj = len(raw_data) - 1
print('n_obj = ', n_obj)
dist = np.zeros((n_obj, n_obj))
for i in range(n_obj):
    for j in range(i + 1):
        if j == i:
            dist[i][i] = 0.0
        else:
            dist[i][j] = float(raw_data[i + 1][j + 1])
            dist[j][i] = dist[i][j]

print('dist...\n', dist)
fout.write('\nDistance Data...\n')
for i in range(n_obj):
    fout.write('{0: <5s}'.format(obj_labels[i]))
    for j in range(i):
        fout.write(' {0: >7.2f}'.format(dist[i][j]))
    fout.write('\n')

n_dim = int(input('Number of dimensions = '))

dist /= dist.max()

#
#       Double centering and characteristic decomposition
#       for initial configuration
#
NegD2 = (dist ** 2) * (-0.5)
J = np.zeros((n_obj, n_obj))
for i in range(n_obj):
    J[i][i] = 1.0
J -= np.ones((n_obj, n_obj)) / n_obj

XXp = J @ NegD2 @ J             #   Double centering
w, v = np.linalg.eigh(XXp)      #   Characteristic decomposition
idx = np.argsort(-w)

w = w[idx]
v = (v.T[idx]).T
X_init = np.zeros((n_obj, n_dim))
for t in range(n_dim):
    if w[t] > 0.0:
        for i in range(n_obj):
            X_init[i][t] = v[i][t] * (w[t] ** 0.5)

x0 = np.zeros(n_obj * n_dim)

d_delta = np.empty((n_obj * (n_obj - 1)) // 2)
t_pos = 0
for i in range(1, n_obj):
    for j in range(i):
        d_delta[t_pos] = dist[i][j]
        t_pos += 1

SltnID = 1
while True:
    iter = 0
    x1 = None
    Stress_1 = None
    dhat_t_1 = dist
    dhat_t_0 = None
    while True:
        print('\niter = ', iter)
        func = linear_func(n_obj, n_dim, dhat_t_1, r)       #   Funciton with r-metric
        func_sup = linear_func_sup(n_obj, n_dim, dhat_t_1)  #   Function with sup-metric
        if iter <= 0:
            if SltnID <= 1:
                x0 = X_init.reshape(n_obj * n_dim)
            else:
                x0 = X_init.reshape(n_obj * n_dim) + \
                    np.random.normal(scale = 0.1, size = n_obj * n_dim)
        else:
            x0 = x1.reshape(n_obj * n_dim) + \
                    np.random.normal(scale = 0.1, size = n_obj * n_dim)

        iter += 1

        #       Approximation by r-metric
        res = so.minimize(func.f, x0, method = 'Nelder-Mead',
                          options = {'xatol': 1.0e-7, 'fatol': 1.0e-14})
        #       Calculation by sup-metric
        res = so.minimize(func_sup.f, res.x, method = 'Nelder-Mead',
                          options = {'xatol': 1.0e-7, 'fatol': 1.0e-14})
        x1 = res.x.reshape((n_obj, n_dim))
        means = x1.mean(axis = 0)
        x1 = x1 - means
        #
        #       Estimated distances by the optimized coordinates
        #
        d_hat = np.empty((n_obj * (n_obj - 1)) // 2)
        i_pos = np.empty((n_obj * (n_obj - 1)) // 2)
        j_pos = np.empty((n_obj * (n_obj - 1)) // 2)
        t_pos = 0
        for i in range(1, n_obj):
            for j in range(i):
                v = 0.0
                for t in range(n_dim):
                    if v < np.abs(x1[i][t] - x1[j][t]):
                        v = np.abs(x1[i][t] - x1[j][t])
                d_hat[t_pos] = v
              
                i_pos[t_pos] = i
                j_pos[t_pos] = j
                t_pos += 1
        #
        #       Transformation by up-and-down blocks algorithm
        #
        d_pair = np.array([(u, v, ipos, jpos) for u, v, ipos, jpos in
                           zip(d_delta, d_hat, i_pos, j_pos)],
                          dtype = ([('delta', np.float), ('dhat', np.float),
                                    ('ipos', np.int), ('jpos', np.int)]))
        idx = np.argsort(d_pair, order = ['delta', 'dhat'])
        dhat = d_pair[idx]['dhat']
        i_pair = d_pair[idx]['ipos']
        j_pair = d_pair[idx]['jpos']

        nonmetric = ba.NonMetricAve( dhat )         
        nonmetric.DoAve()
        dhat_t = [d.v for d in nonmetric.Dist]
        delta = d_pair[idx]['delta']
        dhat_ave = np.zeros((n_obj, n_obj))
        for p in range((n_obj * (n_obj - 1)) // 2):
            dhat_ave[i_pair[p]][j_pair[p]] = nonmetric.Dist[p].v
            dhat_ave[j_pair[p]][i_pair[p]] = nonmetric.Dist[p].v
      
        dhat_ave /= dhat_ave.max()

        dhat_t_0 = dhat_t_1
        dhat_t_1 = dhat_ave         #       The new estimated distances

        diff_d = np.sum((dhat_t_0 - dhat_t_1) ** 2)
        diff_d /= (n_obj * (n_obj - 1)) /2

        if (diff_d < 1.0e-3): 
            print('\niter = {}\n'.format(iter))
            v_num = np.sum((dhat - dhat_t) ** 2)
            v_den = np.sum(dhat ** 2)
            Stress_1 = (v_num / v_den) ** 0.5
            break
              
    print('\nSolution-{}'.format(SltnID))
    print('\nStress-1 = ', Stress_1)
    print('\nX =')
    fout.write('\nSolution-{}\n'.format(SltnID))
    fout.write('\nStress-1 = {}\n'.format(Stress_1))
    fout.write('\nX =\n')
    for i in range(n_obj):
        print('{0: <5s}'.format(obj_labels[i]), end = '')
        fout.write('{0: <5s}'.format(obj_labels[i]))
        for t in range(n_dim):
            print(' {0: >7.3f}'.format(x1[i][t]), end = '')
            fout.write(' {0: >7.3f}'.format(x1[i][t]))
        print()
        fout.write('\n')
    #
    #           Shepard diagram
    #
    plt.plot(d_delta, d_hat, 'bo')
    plt.plot(delta, dhat_t, 'b')
    plt.xlabel('$\delta_{i,j}$', fontsize = 14)
    plt.ylabel('$\hat{d}_{i,j}$', fontsize = 14)
    plt.title('Shepard diagram: Solution-{}'.format(SltnID), fontsize = 16)
    plt.tight_layout(h_pad = 5, w_pad = 5)
    plt.show()
    #
    #           Configuration of objects
    #
    dim1 = 0
    dim2 = 1
    if dim2 >= n_dim:
        dim2 -= 1
    while True:
        plt.figure(figsize = (7,7,))
        x_coord = x1.T[dim1]
        y_coord = x1.T[dim2]
        for px, py, s in zip(x_coord, y_coord, obj_labels):
            plt.text(px, py, s, fontsize = 14)
        plt.plot(x_coord, y_coord, 'bo')
        plt.xlabel('Dim-{}'.format(dim1+1), fontsize = 14)
        plt.ylabel('Dim-{}'.format(dim2+1), fontsize = 14)
        plt.title('Map: Solution-{}'.format(SltnID), fontsize = 16)
        plt.tight_layout(h_pad = 5, w_pad = 5)
        plt.show()

        print('For dimension, put non-positive integer if you want to stop drawing a map')
        dim1 = int(input('dim1 = ')) - 1
        if dim1 < 0:
            break
        if dim1 > n_dim - 1:
            dim1 = n_dim - 1
            
        dim2 = int(input('dim2 = ')) - 1
        if dim2 < 0:
            dim2 = 0
        elif dim2 > n_dim - 1:
            dim2 = n_dim - 1

    ck = input('C(ontinue) or Q(uit) ? ')
    if ck[0] in {'Q', 'q'}:
        break

    SltnID += 1
    
fout.close()
print('\n{} was saved.\n'.format(fout_name))
