import numpy as np

def coord2dist( x, r ):
    """
        Calculate distances from coordinate x
        by r-metric
    """
    n_obj = len(x)
    n_dim = len(x[0])
    dist = np.zeros((n_obj, n_obj))
    for i in range(n_obj):
        for j in range(i):
            if j == i:
                dist[i][i] = 0.0
            else:
                v = 0.0
                for t in range(n_dim):
                    v += (np.abs(x[i][t] - x[j][t])) ** r
                dist[i][j] = v ** (1.0 / r)
                dist[j][i] = dist[i][j]
    return dist

def coord2dist_sup( x ):
    """
            Calculate distances from coordinate x
            by sup-metric
    """
    n_obj = len(x)
    n_dim = len(x[0])
    dist = np.zeros((n_obj, n_obj))
    for i in range(n_obj):
        for j in range(i):
            if j == i:
                dist[i][i] = 0.0
            else:
                v = 0.0
                for t in range(n_dim):
                    if v < np.abs(x[i][t] - x[j][t]):
                        v = np.abs(x[i][t] - x[j][t])
                dist[i][j] = v
                dist[j][i] = dist[i][j]
    return dist

class linear_func:
    """
            Calculate sum of squared differences between
            the target distances and derived r-metric
            distances from flattened coordinate z
    """
    def __init__(self, n_obj, n_dim, dist, r):
        self.n_obj = n_obj
        self.n_dim = n_dim
        self.dist = dist
        self.r = r

    def f(self, z):
        x = z.reshape((self.n_obj, self.n_dim))
        t_dist = coord2dist( x, self.r )
        err = np.sum((self.dist - t_dist)**2)
        return err


class linear_func_sup:
    """
            Calculate sum of squared differences between
            the target distances and derived sup-metric
            distances from flattened coordinate z
    """
    def __init__(self, n_obj, n_dim, dist):
        self.n_obj = n_obj
        self.n_dim = n_dim
        self.dist = dist

    def f(self, z):
        x = z.reshape((self.n_obj, self.n_dim))
        t_dist = coord2dist_sup( x )
        err = np.sum((self.dist - t_dist)**2)
        return err


